"""Defines interface for all job parameters."""

import abc


class JobParametersInterface(abc.ABC):
    """Job Parameters interface definition."""

    DELIMITER = "__"

    @property
    @abc.abstractmethod
    def estimated_run_minutes(self):
        """Return the estimated execution time or None if it isn't known.

        Returns
        -------
        float | None

        """

    @property
    @abc.abstractmethod
    def extension(self):
        """Return the extension name.

        Returns
        -------
        str

        """

    @property
    @abc.abstractmethod
    def name(self):
        """Return the job name. The job name must be unique in a configuration
        and must be usable as a directory or file name on any filesystem.

        Returns
        -------
        str

        """

    @abc.abstractmethod
    def serialize(self):
        """Serialize data to a dictionary.

        Returns
        -------
        dict

        """

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, data):
        """Deserialize parameters from a dictionary.

        Parameters
        ----------
        data : dict

        Returns
        -------
        JobParametersInterface

        """

    @abc.abstractmethod
    def get_blocking_jobs(self):
        """Return the job names blocking this job.

        Returns
        -------
        set
            Empty set means that the job is not blocked.

        """

    @abc.abstractmethod
    def remove_blocking_job(self, name):
        """Remove the name from the job's blocking list.

        Parameters
        ----------
        name : str
            name of job that is now finished

        """

    @abc.abstractmethod
    def set_blocking_jobs(self, blocking_jobs):
        """Set the blocking jobs.

        Parameters
        ----------
        blocking_jobs : set

        """

    @property
    @abc.abstractmethod
    def cancel_on_blocking_job_failure(self):
        """Return False if the job should be canceled if any blocking job fails.

        Returns
        -------
        bool

        """

    @property
    @abc.abstractmethod
    def submission_group(self):
        """Return the submission group for the job."""

    @submission_group.setter
    @abc.abstractmethod
    def submission_group(self, group):
        """Set the submission group for the job.

        Parameters
        ----------
        group : SubmissionGroup

        """

    def is_spark_job(self):
        """Return True if the job is a Spark job.

        Returns
        -------
        bool

        """
        return False



"""Main CLI command for jade."""

from jade.cli.run_spark_cluster import run_spark_cluster
import logging


import click

from jade.cli.run_jobs import run_jobs
from jade.cli.run_multi_node_job import run_multi_node_job
from jade.cli.run_spark_cluster import (
    run_spark_cluster,
    start_spark_cluster_master,
    start_spark_cluster_worker,
)
from jade.cli.run import run


logger = logging.getLogger(__name__)


@click.group()
def cli():
    """Entry point"""


cli.add_command(run)
cli.add_command(run_jobs)
cli.add_command(run_multi_node_job)
cli.add_command(run_spark_cluster)
cli.add_command(start_spark_cluster_master)
cli.add_command(start_spark_cluster_worker)



"""CLI to start a SLURM cluster."""

from jade.jobs.job_configuration_factory import create_config_from_file
import logging
import os
import re
import sys
import time

import click

from jade.common import OUTPUT_DIR
from jade.enums import Status
from jade.jobs.job_runner import JobRunner
from jade.loggers import setup_logging, setup_event_logging
from jade.utils.subprocess_manager import run_command
from jade.utils.utils import get_cli_string


@click.argument(
    "config-file",
    type=str,
)
@click.option(
    "--distributed-submitter/--no-distributed-submitter",
    is_flag=True,
    default=True,
    show_default=True,
    help="Enable distributed submitter",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "-q",
    "--num-parallel-processes-per-node",
    "--num-processes",
    default=None,
    show_default=False,
    type=int,
    help="Number of processes to run in parallel; defaults to num CPUs.",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
@click.command()
def run_jobs(config_file, distributed_submitter, output, num_parallel_processes_per_node, verbose):
    """Starts jobs on HPC."""
    match = re.search(r"batch_(\d+)\.json", config_file)
    assert match
    batch_id = match.group(1)
    os.makedirs(output, exist_ok=True)

    config = create_config_from_file(config_file)
    mgr = JobRunner(config, output=output, batch_id=batch_id)

    # Logging has to get enabled after the JobRunner is created because we need the node ID
    # is what makes the file unique.
    filename = os.path.join(output, f"run_jobs_batch_{batch_id}_{mgr.node_id}.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_event_logging(mgr.event_filename, mode="a")
    logger = setup_logging(__name__, filename, file_level=level, console_level=level, mode="a")
    logger.info(get_cli_string())

    status = mgr.run_jobs(
        distributed_submitter=distributed_submitter,
        verbose=verbose,
        num_parallel_processes_per_node=num_parallel_processes_per_node,
    )
    ret = status.value

    if status == Status.GOOD and distributed_submitter:
        start = time.time()
        _try_submit_jobs(output, verbose=verbose)
        logger.info("try-submit-jobs took %s seconds", time.time() - start)

    sys.exit(ret)


def _try_submit_jobs(output, verbose):
    try_submit_cmd = f"jade try-submit-jobs {output}"
    if verbose:
        try_submit_cmd += " --verbose"
    ret = run_command(try_submit_cmd)
    if ret != 0:
        logger = logging.getLogger(__name__)
        logger.error("Failed to run '%s' ret=%s", try_submit_cmd, ret)



"""CLI to show extensions."""

import logging
import os
import shutil
import sys
from pathlib import Path

import click

from jade.common import HPC_CONFIG_FILE, OUTPUT_DIR
from jade.hpc.common import HpcType
from jade.loggers import setup_logging
from jade.jobs.pipeline_manager import PipelineManager
from jade.models import (
    HpcConfig,
    LocalHpcConfig,
    SingularityParams,
    SubmitterParams,
    get_model_defaults,
)
from jade.utils.utils import get_cli_string, load_data


logger = logging.getLogger(__name__)

SUBMITTER_PARAMS_DEFAULTS = get_model_defaults(SubmitterParams)


@click.group()
def pipeline():
    """Manage JADE execution pipeline."""
    setup_logging("pipeline", None)


@click.command()
@click.option(
    "-a",
    "--auto-config-commands",
    multiple=True,
    help="Commands or scripts to invoke at runtime to create configs for each stage.",
)
@click.option(
    "-f",
    "--config-files",
    multiple=True,
    type=click.Path(exists=True),
    help="Config files, one per stage. Mutually exclusive with --auto-config-cmds.",
)
@click.option(
    "-b",
    "--per-node-batch-size",
    default=SUBMITTER_PARAMS_DEFAULTS["per_node_batch_size"],
    show_default=True,
    help="Number of jobs to run on one node in one batch.",
)
@click.option(
    "-c",
    "--config-file",
    type=click.Path(),
    default=PipelineManager.CONFIG_FILENAME,
    show_default=True,
    help="pipeline config file.",
)
@click.option(
    "-h",
    "--hpc-config",
    type=click.Path(),
    default=HPC_CONFIG_FILE,
    show_default=True,
    help="HPC config file.",
)
@click.option(
    "-l",
    "--local",
    is_flag=True,
    default=False,
    show_default=True,
    help="Run locally even if on HPC.",
)
@click.option(
    "-n",
    "--max-nodes",
    default=None,
    type=int,
    show_default=True,
    help="Max number of node submission requests to make in parallel. Default is unbounded.",
)
@click.option(
    "-p",
    "--poll-interval",
    default=SUBMITTER_PARAMS_DEFAULTS["poll_interval"],
    type=float,
    show_default=True,
    help="Interval in seconds on which to poll jobs for status.",
)
@click.option(
    "-q",
    "--num-parallel-processes-per-node",
    "--num-processes",
    default=None,
    show_default=False,
    type=int,
    help="Number of processes to run in parallel; defaults to num CPUs.",
)
@click.option(
    "--reports/--no-reports",
    is_flag=True,
    default=True,
    show_default=True,
    help="Generate reports after execution.",
)
@click.option(
    "-S",
    "--enable-singularity",
    is_flag=True,
    default=False,
    show_default=True,
    help="Add Singularity parameters and set the config to run in a container.",
)
@click.option(
    "-C",
    "--container",
    type=click.Path(exists=True),
    help="Path to container",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def create(
    config_files,
    auto_config_commands,
    per_node_batch_size,
    config_file,
    hpc_config,
    local,
    max_nodes,
    poll_interval,
    num_parallel_processes_per_node,
    reports,
    enable_singularity,
    container,
    verbose,
):
    """Create a pipeline with multiple Jade configurations. The configs can be specified with
    individual config files or scripts that will be invoked at runtime to create configs based on
    dynamic inputs."""
    if local:
        hpc_config = HpcConfig(hpc_type=HpcType.LOCAL, hpc=LocalHpcConfig())
    else:
        if not os.path.exists(hpc_config):
            print(
                f"{hpc_config} does not exist. Generate it with 'jade config hpc' "
                "or run in local mode with '-l'",
                file=sys.stderr,
            )
            sys.exit(1)
        hpc_config = HpcConfig(**load_data(hpc_config))

    if not config_files and not auto_config_commands:
        print("Either --config-files or --auto-config-commands must be specified", file=sys.stderr)
        sys.exit(1)
    if config_files and auto_config_commands:
        print(
            "You cannot specify both --config-files and  --auto-config-commands", file=sys.stderr
        )
        sys.exit(1)

    if enable_singularity:
        singularity_params = SingularityParams(enabled=True, container=container)
    else:
        singularity_params = None
    submit_params = SubmitterParams(
        generate_reports=reports,
        hpc_config=hpc_config,
        max_nodes=max_nodes,
        num_parallel_processes_per_node=num_parallel_processes_per_node,
        per_node_batch_size=per_node_batch_size,
        poll_interval=poll_interval,
        singularity_params=singularity_params,
        verbose=verbose,
    )
    if config_files:
        PipelineManager.create_config_from_files(config_files, config_file, submit_params)
    else:
        PipelineManager.create_config_from_commands(
            auto_config_commands, config_file, submit_params
        )


@click.command()
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def status(output, verbose):
    """Check status of the pipeline."""
    try:
        mgr = PipelineManager.load(output)
    except FileNotFoundError:
        print(f"{output} is not a valid pipeline output directory", file=sys.stderr)
        sys.exit(1)

    config = mgr.config
    completed_stages = []
    current_stage = None
    for stage in config.stages:
        if stage.stage_num < config.stage_num:
            completed_stages.append(stage)
        elif stage.stage_num == config.stage_num:
            current_stage = stage

    print(f"Is complete: {config.is_complete}")
    if current_stage is not None:
        print(f"Current stage number: {config.stage_num}")
        print("\nTo view the status of the current stage:")
        print(f"  jade show-status -o {current_stage.path}")
    if completed_stages:
        print(f"\nTo view results of the completed stages:")
        for stage in completed_stages:
            print(f"  jade show-results -o {stage.path}")


@click.command()
@click.argument("config-file")
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    show_default=True,
    help="Delete output directory if it exists.",
)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def submit(config_file, output, force, verbose=False):
    """Submit the pipeline for execution."""
    if os.path.exists(output):
        if force:
            shutil.rmtree(output)
        else:
            print(
                f"{output} already exists. Delete it or use '--force' to overwrite.",
                file=sys.stderr,
            )
            sys.exit(1)
    os.makedirs(output, exist_ok=True)

    filename = os.path.join(output, "pipeline_submit.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level)
    logger.info(get_cli_string())

    mgr = PipelineManager.create(config_file, output)
    try:
        mgr.submit_next_stage(1)
    except Exception:
        logger.exception("Pipeline execution failed")
        raise

    logging.shutdown()
    sys.exit(0)


@click.command()
@click.argument("output")
@click.option(
    "--stage-num",
    required=True,
    type=int,
    help="stage number to submit",
)
@click.option(
    "--return-code",
    required=True,
    type=int,
    help="return code of stage index that just completed",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def submit_next_stage(output, stage_num, return_code, verbose=False):
    """Internal command to submit the next stage of the pipeline for execution."""
    filename = os.path.join(output, "pipeline_submit.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="a")
    logger.info(get_cli_string())

    mgr = PipelineManager.load(output)
    try:
        mgr.submit_next_stage(stage_num, return_code=return_code)
    except Exception:
        logger.exception("Pipeline execution failed")
        raise

    logging.shutdown()
    sys.exit(0)


pipeline.add_command(create)
pipeline.add_command(status)
pipeline.add_command(submit)
pipeline.add_command(submit_next_stage)



"""Generic command CLI"""

import os

from jade.extensions.generic_command import GenericCommandConfiguration


def auto_config(inputs, **kwargs):
    """Create a configuration file for generic_command.

    Parameters
    ----------
    inputs : str
        Input file containing commands, one line per command

    """
    if not os.path.exists(inputs):
        raise OSError(f"Inputs path '{inputs}' does not exist.")

    return GenericCommandConfiguration.auto_config(inputs, **kwargs)


def run(config_file, name, output, output_format, verbose):
    """Run auto regression analysis through command line"""
    assert False



"""Manages execution of jobs on a node."""

import logging
import os
import shutil
import time
import uuid
import time
from pathlib import Path

from jade.common import JOBS_OUTPUT_DIR, OUTPUT_DIR
from jade.enums import Status, ResourceMonitorType
from jade.exceptions import ExecutionError
from jade.hpc.common import HpcType
from jade.hpc.hpc_manager import HpcManager
from jade.jobs.cluster import Cluster
from jade.jobs.async_cli_command import AsyncCliCommand
from jade.jobs.job_manager_base import JobManagerBase
from jade.jobs.job_queue import JobQueue
from jade.models import JobState
from jade.loggers import close_event_logging
from jade.resource_monitor import ResourceMonitorAggregator, ResourceMonitorLogger
from jade.utils.run_command import run_command, check_run_command
from jade.utils.timing_utils import timed_info


logger = logging.getLogger(__name__)


class JobRunner(JobManagerBase):
    """Manages execution of jobs on a node."""

    def __init__(
        self,
        config,
        output,
        batch_id=0,
    ):
        super(JobRunner, self).__init__(config, output)
        self._handle_submission_groups()
        group = self.config.get_default_submission_group()
        hpc_config = group.submitter_params.hpc_config
        self._intf = HpcManager.create_hpc_interface(hpc_config)
        self._node_id = self._intf.get_node_id()
        self._intf_type = hpc_config.hpc_type
        self._batch_id = batch_id
        self._event_filename = os.path.join(
            output,
            f"run_jobs_batch_{batch_id}_{self._node_id}_events.log",
        )
        self._event_logger = None

        logger.debug("Constructed JobRunner output=%s batch=%s", output, batch_id)

    @property
    def event_filename(self):
        return self._event_filename

    @property
    def node_id(self):
        """Return the node ID of the current system.

        Returns
        -------
        str

        """
        return self._node_id

    @timed_info
    def run_jobs(
        self, distributed_submitter=True, verbose=False, num_parallel_processes_per_node=None
    ):
        """Run the jobs.

        Parameters
        ----------
        distributed_submitter : bool
            If True, make cluster updates.
        verbose : bool
            If True, enable debug logging.
        num_parallel_processes_per_node : int
            Number of processes to run in parallel; defaults to num CPUs

        Returns
        -------
        Status

        """
        logger.info("Run jobs.")
        scratch_dir = self._create_local_scratch()
        are_inputs_local = self._intf_type == HpcType.LOCAL

        try:
            config_file = self._config.serialize_for_execution(scratch_dir, are_inputs_local)
            jobs = self._generate_jobs(config_file, verbose)

            env = os.environ.copy()
            env["JADE_RUNTIME_OUTPUT"] = str(self._output)
            env["JADE_SUBMISSION_GROUP"] = self._config.get_default_submission_group().name
            # Setting node_setup_script and node_shutdown_script are obsolete and will
            # eventually be deleted.
            group = self._config.get_default_submission_group()
            if group.submitter_params.node_setup_script is not None:
                cmd = f"{group.submitter_params.node_setup_script} {config_file} {self._output}"
                check_run_command(cmd, env=env)
            elif self._config.node_setup_command is not None:
                check_run_command(self._config.node_setup_command, env=env)

            result = self._run_jobs(
                jobs, num_parallel_processes_per_node=num_parallel_processes_per_node
            )

            if group.submitter_params.node_shutdown_script:
                cmd = f"{group.submitter_params.node_shutdown_script} {config_file} {self._output}"
                ret2 = run_command(cmd, env=env)
                if ret2 != 0:
                    logger.error("Failed to run node shutdown script %s: %s", cmd, ret2)
            elif self._config.node_teardown_command is not None:
                start = time.time()
                ret2 = run_command(self._config.node_teardown_script, env=env)
                if ret2 != 0:
                    logger.error(
                        "Failed to run node shutdown script %s: %s",
                        self._config.node_teardown_command,
                        ret2,
                    )
                logger.info("Node teardown script duration = %s seconds", time.time() - start)

            logger.info("Completed %s jobs", len(jobs))
        finally:
            shutil.rmtree(scratch_dir)
            if distributed_submitter and are_inputs_local:
                self._complete_hpc_job()

        return result

    def _complete_hpc_job(self):
        """Complete the HPC job in the cluster. A future submitter could detect this. However,
        it's better to do it here for these reasons:

        1. If we don't do this and the user runs 'jade show-status -j' before the next submitter
           runs, a stale HPC job ID will show up.
        2. The last submitter will never detect its job as complete, and so the job status file
           will always include it.

        """
        job_id = self._intf.get_current_job_id()
        if job_id is None:
            # TODO: need to implement persistent recording of fake status for FakeManager
            return

        max_time_s = 30
        interval = 5
        completed = False
        for _ in range(max_time_s // interval):
            cluster, promoted = Cluster.deserialize(
                self._output, try_promote_to_submitter=True, deserialize_jobs=True
            )
            if promoted:
                try:
                    cluster.complete_hpc_job_id(job_id)
                finally:
                    cluster.demote_from_submitter()
                completed = True
                break
            time.sleep(interval)

        if not completed:
            logger.warning(
                "Could not promote to submitter. HPC job ID %s is still present", job_id
            )
            # Couldn't acquire lock because other nodes were continually doing work (unexpected).
            # A future submitter will see that this job ID is complete and clear the condition.

    def _create_local_scratch(self):
        local_scratch = self._intf.get_local_scratch()
        dirname = "jade-" + str(uuid.uuid4())
        scratch_dir = os.path.join(local_scratch, dirname)
        os.makedirs(scratch_dir, exist_ok=True)
        logger.info("Created jade scratch_dir=%s", scratch_dir)
        return scratch_dir

    def _generate_jobs(self, config_file, verbose):
        jobs = []
        for job in self._config.iter_jobs():
            job_exec_class = self._config.job_execution_class(job.extension)
            djob = AsyncCliCommand(
                job,
                job_exec_class.generate_command(
                    job,
                    self._jobs_output,
                    config_file,
                    verbose=verbose,
                ),
                self._output,
                self._batch_id,
                self._intf.am_i_manager(),
                self._intf.get_current_job_id(),
            )
            jobs.append(djob)

        return jobs

    def _run_jobs(self, jobs, num_parallel_processes_per_node=None):
        num_jobs = len(jobs)
        if num_parallel_processes_per_node is None:
            max_num_workers = self._intf.get_num_cpus()
        else:
            max_num_workers = num_parallel_processes_per_node
        num_workers = min(num_jobs, max_num_workers)
        logger.info(
            "Generated %s jobs to execute on %s workers max=%s.",
            num_jobs,
            num_workers,
            max_num_workers,
        )
        self._intf.log_environment_variables()

        name = f"resource_monitor_batch_{self._batch_id}_{self._node_id}"
        group = self._config.get_default_submission_group()
        monitor_type = group.submitter_params.resource_monitor_type
        resource_aggregator = None
        if monitor_type == ResourceMonitorType.AGGREGATION:
            resource_aggregator = ResourceMonitorAggregator(
                name, group.submitter_params.resource_monitor_stats
            )
            monitor_func = resource_aggregator.update_resource_stats
            resource_monitor_interval = group.submitter_params.resource_monitor_interval
        elif monitor_type == ResourceMonitorType.PERIODIC:
            resource_monitor = ResourceMonitorLogger(
                name, group.submitter_params.resource_monitor_stats
            )
            monitor_func = resource_monitor.log_resource_stats
            resource_monitor_interval = group.submitter_params.resource_monitor_interval
        elif monitor_type == ResourceMonitorType.NONE:
            monitor_func = None
            resource_monitor_interval = None
        else:
            assert False, monitor_type
        JobQueue.run_jobs(
            jobs,
            max_queue_depth=num_workers,
            monitor_func=monitor_func,
            monitor_interval=resource_monitor_interval,
            poll_interval=group.submitter_params.poll_interval,
        )

        logger.info("Jobs are complete. count=%s", num_jobs)
        if resource_aggregator is not None:
            resource_aggregator.finalize(self._output)
        self._aggregate_events()
        return Status.GOOD  # TODO

    @timed_info
    def _aggregate_events(self):
        # Aggregate all job events.log files into this node's log file so
        # that the we can more quickly make events.json later.
        close_event_logging()
        with open(self._event_filename, "a") as f_out:
            for job in self._config.iter_jobs():
                job_file = os.path.join(self._output, JOBS_OUTPUT_DIR, job.name, "events.log")
                if not os.path.exists(job_file):
                    # Extensions aren't required to create these.
                    continue
                with open(job_file) as f_in:
                    for line in f_in:
                        f_out.write(line)
                os.remove(job_file)
                logger.debug("Moved contents of %s to %s", job_file, self._event_filename)



"""
CLI to show events of a scenario.
"""

import logging
import sys

import click

from jade.common import OUTPUT_DIR
from jade.loggers import setup_logging
from jade.events import EventsSummary


@click.argument("names", nargs=-1)
@click.option("-o", "--output", default=OUTPUT_DIR, show_default=True, help="Output directory.")
@click.option(
    "-c",
    "--categories",
    is_flag=True,
    default=False,
    show_default=True,
    help="Treat the arguments as categories instead of names.",
)
@click.option(
    "-j",
    "--json-fmt",
    is_flag=True,
    default=False,
    show_default=True,
    help="Print event in JSON format, instead of table",
)
@click.option(
    "-n",
    "--names-only",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show event names in output.",
)
@click.option(
    "-C",
    "--categories-only",
    is_flag=True,
    default=False,
    show_default=True,
    help="Show event categories in output.",
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable verbose event outputs.",
)
@click.command()
def show_events(
    output,
    names,
    categories=False,
    json_fmt=False,
    names_only=False,
    categories_only=False,
    verbose=False,
):
    """Shows the events after jobs run.

    \b
    Examples:
    jade show-events
    jade show-events unhandled_error
    jade show-events error -c
    jade show-events --names-only
    jade show-events --categories-only
    """
    level = logging.DEBUG if verbose else logging.WARNING
    setup_logging("show_results", None, console_level=level)
    results = EventsSummary(output)
    if names_only:
        results.show_event_names()
    elif categories_only:
        results.show_event_categories()
    elif json_fmt:
        print(results.to_json())
    else:
        if not names:
            names = results.list_unique_names()
        if not names:
            print(f"There are no events in {output}", file=sys.stderr)
        for name in names:
            if categories:
                results.show_events_in_category(name)
            else:
                results.show_events(name)



"""
The job execution class and methods for auto-regression
"""
import os

import numpy as np
import pandas as pd
from statsmodels.tsa.ar_model import AR
import matplotlib

from jade.events import StructuredErrorLogEvent, EVENT_CATEGORY_ERROR, EVENT_NAME_UNHANDLED_ERROR
from jade.jobs.job_execution_interface import JobExecutionInterface
from jade.loggers import log_event
from jade.utils.utils import dump_data


def autoregression_analysis(country, data, output):
    """
    Country based GDP auto-regression analysis

    Parameters
    ----------
    country: str
        the name of a country
    data: str
        path to the csv file containing the GDP data.
    output: str
        The path to the output directory

    Returns
    -------
    tuple, The path of csv result file, and the path of png plot file.
    """
    # Read csv
    df = pd.read_csv(data, index_col="year")
    df = df.dropna()

    # Train model
    train = df["gdp"].values
    model = AR(train)
    model_fit = model.fit()

    # Validate model
    lag = model_fit.k_ar
    pred = model_fit.predict(start=lag, end=len(train), dynamic=False)

    # Save result
    df["pred_gdp"] = [np.nan for _ in range(lag - 1)] + list(pred)
    result_file = os.path.join(output, "result.csv")
    df.to_csv(result_file)

    # Save plot
    matplotlib.use("Agg")
    import matplotlib.pyplot as plt

    df.plot()
    plt.grid(axis="y", linestyle="--")
    plt.title(country + "(current $)")
    plot_file = os.path.join(output, "result.png")
    plt.savefig(plot_file)

    return result_file, plot_file


class AutoRegressionExecution(JobExecutionInterface):
    """
    A class used for auto-regression job execution on computer.
    """

    def __init__(self, job, output, output_format="csv"):
        """
        Init auto-regression execution class

        Parameters
        ----------
        job: :obj:`AutoRegressionParameters`
            The instance of :obj:`AutoRegressionParameters`
        output: str,
            The path to the output directory.
        output_format: str,
            The export format of result file, default 'csv'.
        """
        self._job = job
        self._output = output
        self._output_format = output_format
        self._job_dir = os.path.join(output, job.name)
        os.makedirs(self._job_dir, exist_ok=True)

    @property
    def results_directory(self):
        """Return the results directory created by the simulation."""
        return self._output

    @classmethod
    def create(cls, _, job, output, **kwargs):
        """Create instance of :obj:`AutoRegressionExecution`"""
        return cls(job, output)

    @staticmethod
    def generate_command(job, output, config_file, verbose=False):
        """
        Generate command consumed by bash for running auto-regression analysis.

        Parameters
        ----------
        job: :obj:`AutoRegressionParameters`
            The instance of :obj:`AutoRegressionParameters`.
        output: str
            The path to the output directory.
        config_file: str,
            The path to the configuration json file of job inputs.
        verbose: bool
            True if verbose, otherwise False.

        Returns
        -------
            str, A command line string
        """
        extension = "demo"
        command = [
            f"jade-internal run {extension}",
            f"--name={job.name}",
            f"--output={output}",
            f"--config-file={config_file}",
        ]

        if verbose:
            command.append("--verbose")

        return " ".join(command)

    def list_results_files(self):
        """Return a list of result filenames created by the simulation."""
        return [os.path.join(self._output, x) for x in os.listdir(self._output)]

    def post_process(self, **kwargs):
        """Run post-process operations on data."""

    def run(self):
        """Runs the autoregression, and return status code"""
        try:
            result_file, plot_file = autoregression_analysis(
                country=self._job.country, data=self._job.data, output=self._job_dir
            )
            summary_data = {
                "name": self._job.name,
                "country": self._job.country,
                "output": self._output,
                "result": result_file,
                "plot": plot_file,
            }
            summary_file = os.path.join(self._job_dir, "summary.toml")
            dump_data(summary_data, summary_file)
            if self._job.country == "australia":
                raise Exception("test")

        # Log event into file
        except Exception:
            # Create event instance
            event = StructuredErrorLogEvent(
                source=self._job.name,
                category=EVENT_CATEGORY_ERROR,
                name=EVENT_NAME_UNHANDLED_ERROR,
                message="Analysis failed!",
            )

            # Log event into file with structured message.
            log_event(event)

            # Must raise the exception here, or job returncode is 0 even it fails.
            raise

        return 0



"""Models for pipeline execution"""

from typing import List, Optional

from pydantic.v1 import Field

from jade.models import JadeBaseModel, SubmitterParams


class PipelineStage(JadeBaseModel):
    """Describes one stage of a pipeline."""

    auto_config_cmd: Optional[str] = Field(
        title="auto_config_cmd",
        description="command used to create the JADE configuration",
    )
    config_file: str = Field(
        title="config_file",
        description="JADE configuration file",
    )
    stage_num: int = Field(
        title="stage_num",
        description="1-based ID of the stage in the pipeline",
    )
    path: Optional[str] = Field(
        title="path",
        description="directory on shared filesystem containing config",
    )
    return_code: Optional[int] = Field(
        title="return_code",
        description="return code of stage; 0 is success",
    )
    submitter_params: SubmitterParams = Field(
        title="submitter_params",
        description="defines the submitter params selected by the user",
    )


class PipelineConfig(JadeBaseModel):
    """Describes the configuration and status of a pipeline."""

    path: Optional[str] = Field(
        title="path",
        description="directory on shared filesystem containing config",
    )
    stage_num: int = Field(
        title="stage_num",
        description="number of current stage",
    )
    stages: List[PipelineStage] = Field(
        title="stages",
        description="stages in the pipeline",
    )
    is_complete: bool = Field(
        title="is_complete",
        description="set to True when the pipeline is complete",
        default=False,
    )



"""SLURM management functionality"""

import logging
import multiprocessing
import tempfile
import time
from pathlib import Path

from filelock import SoftFileLock

from jade.enums import Status
from jade.hpc.common import HpcJobStatus, HpcJobInfo
from jade.hpc.hpc_manager_interface import HpcManagerInterface
from jade.utils.subprocess_manager import SubprocessManager
from jade.utils.utils import create_script


logger = logging.getLogger(__name__)


DEFAULTS = {
    "walltime": 60 * 12,
    "interface": "ib0",
    "local_directory": tempfile.gettempdir(),
    "memory": 5000,
}


class FakeManager(HpcManagerInterface):
    """Simulates management of HPC jobs."""

    JOB_ID_FILE = "fake_manager_job_id.txt"
    LOCK_FILE = "fake_manager.lock"

    def __init__(self, config):
        self._subprocess_mgr = None
        self._job_id = None
        self._config = config

    def am_i_manager(self):
        return True

    def cancel_job(self, job_id):
        return 0

    def check_status(self, name=None, job_id=None):
        if self._subprocess_mgr is None:
            job_info = HpcJobInfo(job_id, "", HpcJobStatus.NONE)
        elif self._subprocess_mgr.in_progress():
            job_info = HpcJobInfo(job_id, "", HpcJobStatus.RUNNING)
        else:
            job_info = HpcJobInfo(job_id, "", HpcJobStatus.COMPLETE)

        logger.debug("status=%s", job_info)
        return job_info

    def check_statuses(self):
        val = {self._job_id: self.check_status(job_id=self._job_id).status}
        return val

    def check_storage_configuration(self):
        pass

    def create_cluster(self):
        pass

    def create_local_cluster(self):
        pass

    def create_submission_script(self, name, script, filename, path):
        lines = [
            "#!/bin/bash",
            str(script),
        ]
        create_script(filename, "\n".join(lines))

    def get_config(self):
        return {"hpc": {}}

    def get_current_job_id(self):
        return None

    def get_job_stats(self, job_id):
        return None

    def get_job_stats(self, job_id):
        return None

    def get_local_scratch(self):
        return tempfile.gettempdir()

    def get_node_id(self):
        # If we try to use multi-node jobs in fake mode, this will cause a problem.
        return "0"

    @staticmethod
    def get_num_cpus():
        return multiprocessing.cpu_count()

    def list_active_nodes(self, job_id):
        assert False

    def log_environment_variables(self):
        pass

    @staticmethod
    def _get_next_job_id(output_path):
        """Returns the next job ID and increments the index.
        A lock must be held while calling this method.

        """
        # TODO: This could be enhanced to record completions.
        path = output_path / FakeManager.JOB_ID_FILE
        if path.exists():
            job_id = int(path.read_text().strip())
        else:
            job_id = 1
        next_job_id = job_id + 1
        path.write_text(str(next_job_id) + "\n")
        return job_id

    def submit(self, filename):
        # This method has a workaround for problems seen on some Linux systems
        # (never on Mac).
        # When multiple processes call this method at about the same time,
        # one or more of the subprocesses do not get started. It seems like
        # something within Python gets locked up.
        # This workaround staggers starting of the subprocesses and prevents
        # the issue from occurring.
        output_path = Path(filename).parent
        with SoftFileLock(output_path / FakeManager.LOCK_FILE, timeout=30):
            self._job_id = self._get_next_job_id(output_path)
            self._subprocess_mgr = SubprocessManager()
            self._subprocess_mgr.run(filename)
            logger.info("Submit job with %s", self._job_id)
            time.sleep(1)
            return Status.GOOD, self._job_id, None



"""Defines an async CLI command."""

import logging
import os
import shlex
import subprocess
import sys
import time
from pathlib import Path

from jade.common import JOBS_OUTPUT_DIR, JOBS_STDIO_DIR, RESULTS_DIR
from jade.enums import JobCompletionStatus, Status
from jade.events import StructuredLogEvent, EVENT_NAME_BYTES_CONSUMED, EVENT_CATEGORY_RESOURCE_UTIL
from jade.jobs.async_job_interface import AsyncJobInterface
from jade.jobs.results_aggregator import ResultsAggregator
from jade.loggers import log_event
from jade.result import Result
from jade.utils.utils import get_directory_size_bytes


logger = logging.getLogger(__name__)


class AsyncCliCommand(AsyncJobInterface):
    """Defines a a CLI command that can be submitted asynchronously."""

    def __init__(self, job, cmd, output, batch_id, is_manager_node, hpc_job_id):
        self._job = job
        self._cli_cmd = cmd
        self._output = Path(output)
        self._pipe = None
        self._is_pending = False
        self._start_time = None
        self._return_code = None
        self._is_complete = False
        self._batch_id = batch_id
        self._is_manager_node = is_manager_node
        self._hpc_job_id = hpc_job_id
        self._stdout_fp = None
        self._stderr_fp = None

    def __del__(self):
        if self._is_pending:
            logger.warning("job %s destructed while pending", self._cli_cmd)

    def _complete(self):
        self._return_code = self._pipe.returncode
        self._stdout_fp.close()
        self._stderr_fp.close()
        exec_time_s = time.time() - self._start_time

        if not self._is_manager_node:
            # This will happen on a multi-node job. Don't complete it multiple times.
            logger.info(
                "Job %s completed on non-manager node return_code=%s exec_time_s=%s",
                self._job.name,
                self._return_code,
                exec_time_s,
            )
            return

        status = JobCompletionStatus.FINISHED
        output_dir = self._output / JOBS_OUTPUT_DIR / self._job.name
        bytes_consumed = get_directory_size_bytes(output_dir)
        event = StructuredLogEvent(
            source=self._job.name,
            category=EVENT_CATEGORY_RESOURCE_UTIL,
            name=EVENT_NAME_BYTES_CONSUMED,
            message="job output directory size",
            bytes_consumed=bytes_consumed,
        )
        log_event(event)
        result = Result(
            self._job.name, self._return_code, status, exec_time_s, hpc_job_id=self._hpc_job_id
        )
        ResultsAggregator.append(self._output, result, batch_id=self._batch_id)

        logger.info(
            "Job %s completed return_code=%s exec_time_s=%s hpc_job_id=%s",
            self._job.name,
            self._return_code,
            exec_time_s,
            self._hpc_job_id,
        )

    def cancel(self):
        self._return_code = 1
        self._is_complete = True
        if self._is_manager_node:
            result = Result(
                self._job.name,
                self._return_code,
                JobCompletionStatus.CANCELED,
                0.0,
                hpc_job_id=self._hpc_job_id,
            )
            ResultsAggregator.append(self._output, result, batch_id=self._batch_id)
            logger.info("Canceled job %s", self._job.name)
        else:
            logger.info("Canceled job %s on non-manager node", self._job.name)

    @property
    def cancel_on_blocking_job_failure(self):
        return self._job.cancel_on_blocking_job_failure

    def get_id(self):
        return self._pipe.pid

    def set_blocking_jobs(self, jobs):
        self._job.set_blocking_jobs(jobs)

    def is_complete(self):
        if self._is_complete:
            return True

        if not self._is_pending:
            ret = self._pipe.poll()
            assert ret is None, f"{ret}"
            return True

        if self._pipe.poll() is not None:
            self._is_pending = False
            self._complete()

        return not self._is_pending

    @property
    def job(self):
        """Get the job.

        Parameters
        ----------
        job : JobParametersInterface

        """
        return self._job

    @property
    def name(self):
        return self._job.name

    def get_blocking_jobs(self):
        return self._job.get_blocking_jobs()

    def remove_blocking_job(self, name):
        self._job.remove_blocking_job(name)

    @property
    def return_code(self):
        return self._return_code

    def run(self):
        """Run the job. Writes results to file when complete."""
        assert self._pipe is None
        self._start_time = time.time()

        # Disable posix if on Windows.
        cmd = shlex.split(self._cli_cmd, posix="win" not in sys.platform)
        env = os.environ.copy()
        env["JADE_RUNTIME_OUTPUT"] = str(self._output)
        env["JADE_JOB_NAME"] = self.name
        stdout_filename = self._output / JOBS_STDIO_DIR / f"{self._job.name}.o"
        stderr_filename = self._output / JOBS_STDIO_DIR / f"{self._job.name}.e"
        self._stdout_fp = open(stdout_filename, "w")
        self._stderr_fp = open(stderr_filename, "w")
        self._pipe = subprocess.Popen(cmd, env=env, stdout=self._stdout_fp, stderr=self._stderr_fp)
        self._is_pending = True
        logger.info("Started job name=%s hpc_job_id=%s", self._job.name, self._hpc_job_id)
        logger.debug("Submitted %s", self._cli_cmd)
        return Status.GOOD



"""Defines base class for job execution analysis jobs."""

import os

from jade.common import ANALYSIS_DIR
from jade.jobs.job_execution_interface import JobExecutionInterface


class AnalysisExecutionBase(JobExecutionInterface):
    """Base class for analysis jobs. This job type is intended to run on the
    output results of another job."""

    def __init__(self, output_dir, simulations_dir):
        self._analysis_dir = self.get_analysis_dir(output_dir)
        self._simulations_dir = simulations_dir
        os.makedirs(self._analysis_dir, exist_ok=True)

    @staticmethod
    def get_analysis_dir(output_dir):
        """Get the analysis directory to use.

        Parameters
        ----------
        output_dir : str

        Returns
        -------
        str

        """
        return os.path.join(output_dir, ANALYSIS_DIR)

    @property
    def results_directory(self):
        """Return the results directory created by the simulation."""
        return self._analysis_dir

    def list_results_files(self):
        """Return a list of result filenames created by the simulation."""
        return [os.path.join(self._analysis_dir, x) for x in os.listdir(self._analysis_dir)]

    def post_process(self, **kwargs):
        pass



"""Collects information about the source code repository."""

import logging
import os
import re

from jade.exceptions import ExecutionError, InvalidParameter
from jade.utils.subprocess_manager import run_command


logger = logging.getLogger(__name__)


class RepositoryInfo:
    """Collects information about the source code repository for a package."""

    def __init__(self, package):
        # This will be the directory containing the package.
        self._path = os.path.dirname(getattr(package, "__path__")[0])

        if not os.path.exists(os.path.join(self._path, ".git")):
            raise InvalidParameter("{package} is not in a git repository")

        self._patch_filename = None

    def _run_command(self, cmd):
        orig = os.getcwd()
        os.chdir(self._path)

        try:
            output = {}
            ret = run_command(cmd, output=output)
            if ret != 0:
                raise ExecutionError(f"[{cmd}] failed: {ret}: {output['stderr']}")

            return output["stdout"].strip()
        finally:
            os.chdir(orig)

    def current_branch(self):
        """Return the current branch.

        Returns
        -------
        str

        """
        cmd = "git rev-parse --abbrev-ref HEAD"
        output = self._run_command(cmd)
        return output

    def last_commit(self):
        """Return the last commit.

        Returns
        -------
        str

        """
        output = self._run_command("git log -n 1")

        regex = re.compile(r"^commit (\w+)")
        match = regex.search(output)
        assert match, output
        commit = match.group(1)

        return commit

    def status(self):
        """Return the current status.

        Returns
        -------
        str

        """
        cmd = "git status --porcelain=v2 --branch --verbose " "--untracked-files=no"
        output = self._run_command(cmd)
        return output

    def write_diff_patch(self, filename):
        """Write any repo diffs to filename.

        Parameters
        ----------
        filename : str

        """
        output = self._run_command("git diff")
        if output:
            with open(filename, "w") as f_out:
                f_out.write(output)
                logger.info("Wrote diff to %s", filename)
                self._patch_filename = filename
        else:
            logger.info("No diff detected in repository.")

    def summary(self):
        """Return a summary of the repository status.

        Returns
        -------
        dict

        """
        return {
            "current_branch": self.current_branch(),
            "diff_patch_file": self._patch_filename,
            "last_commit": self.last_commit(),
            "status": self.status(),
        }



"""Implements the JobParametersInterface for generic_command."""

import logging
from collections import namedtuple
from pathlib import Path
from typing import Dict, List, Optional, Set

from pydantic.v1 import Field, validator

from jade.models import JadeBaseModel
from jade.models.spark import SparkConfigModel, SparkContainerModel
from jade.common import DEFAULT_SUBMISSION_GROUP
from jade.jobs.job_parameters_interface import JobParametersInterface


logger = logging.getLogger(__name__)

_EXTENSION = "generic_command"


class GenericCommandParameters(JobParametersInterface):
    """A class used for creating a job for a generic command."""

    parameters_type = namedtuple("GenericCommand", "command")

    def __init__(self, **kwargs):
        self._model = GenericCommandParametersModel(**kwargs)

    def __str__(self):
        return "<GenericCommandParameters: {}>".format(self.name)

    def __getattr__(self, name):
        if name in GenericCommandParametersModel.__fields__:
            return getattr(self._model, name)
        raise AttributeError(f"'GenericCommandParameters' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name == "extension":
            raise AttributeError(f"'GenericCommandParameters' does not allow setting 'extension'")
        if name in GenericCommandParametersModel.__fields__:
            setattr(self._model, name, value)
        self.__dict__[name] = value

    @property
    def command(self):
        if self._model.use_multi_node_manager:
            return f"jade-internal run-multi-node-job {self.name} {self._model.command}"
        elif self.is_spark_job():
            return f"jade-internal run-spark-cluster {self.name} {self._model.command}"
        return self._model.command

    @property
    def estimated_run_minutes(self):
        return self._model.estimated_run_minutes

    @property
    def extension(self):
        return _EXTENSION

    @property
    def name(self):
        return self._create_name() if self._model.name is None else self._model.name

    def _create_name(self):
        return str(self._model.job_id)

    def serialize(self):
        assert self._model.job_id is not None
        # If job sizes get huge then we should exclude parameters with default values.
        return self._model.dict()

    @classmethod
    def deserialize(cls, data):
        return cls(**data)

    @property
    def cancel_on_blocking_job_failure(self):
        return self._model.cancel_on_blocking_job_failure

    def get_blocking_jobs(self):
        return self._model.blocked_by

    @property
    def model(self):
        return self._model

    def remove_blocking_job(self, name):
        self._model.blocked_by.remove(name)

    def set_blocking_jobs(self, blocking_jobs):
        self._model.blocked_by = blocking_jobs

    @property
    def submission_group(self):
        return self._model.submission_group

    def is_spark_job(self):
        return self._model.spark_config is not None and self._model.spark_config.enabled


class GenericCommandParametersModel(JadeBaseModel):
    """Model definition for generic command parameters"""

    name: Optional[str] = Field(
        title="name",
        description="If not set Jade will use the job_id converted to a string. Must be unique.",
    )
    use_multi_node_manager: bool = Field(
        title="use_multi_node_manager",
        description="If true JADE will run this job with its multi-node manager.",
        default=False,
    )
    spark_config: Optional[SparkConfigModel] = Field(
        title="spark_config",
        description="If enabled JADE will run this job on a Spark cluster.",
        default=None,
    )
    command: str = Field(
        title="command",
        description="Command that can be invoked in a terminal (shell characters not allowed)",
    )
    blocked_by: Set[str] = Field(
        title="blocked_by",
        description="Array of job names that must complete before this one can start.",
        default=set(),
    )
    cancel_on_blocking_job_failure: bool = Field(
        title="cancel_on_blocking_job_failure",
        description="If true JADE will cancel this job if any of its blocking jobs fail.",
        default=False,
    )
    estimated_run_minutes: Optional[int] = Field(
        title="estimated_run_minutes",
        description="JADE will use this value along with num-parallel-processes-per-node and "
        "walltime to build per-node batches of jobs if time-based-batching is enabled.",
    )
    submission_group: str = Field(
        title="submission_group",
        description="Optional name of a submission group",
        default=DEFAULT_SUBMISSION_GROUP,
    )
    append_job_name: bool = Field(
        title="append_job_name",
        description="If true JADE will append --jade-job-name=X where X is the job's name.",
        default=False,
    )
    append_output_dir: bool = Field(
        title="append_output_dir",
        description="If true JADE will append --jade-runtime-output=X where X is the output "
        "directory specified in jade submit-jobs.",
        default=False,
    )
    ext: Dict = Field(
        title="ext",
        description="User-defined extension data to be used at runtime. Must be serializable in "
        "JSON format.",
        default={},
    )
    job_id: Optional[int] = Field(
        title="job_id",
        description="Unique job identifier, generated by JADE",
    )
    extension: str = Field(
        title="extension",
        description="job extension type, generated by JADE",
        default=_EXTENSION,
    )

    @validator("append_output_dir")
    def handle_append_output_dir(cls, value, values):
        spark_enabled = False
        if values["spark_config"] is not None:
            spark_enabled = getattr(values["spark_config"], "enabled")
        if values["use_multi_node_manager"] or spark_enabled:
            logger.debug(
                "Override 'append_output_dir' because 'use_multi_node_manager' is set or spark is enabled"
            )
            return True
        return value

    @validator("blocked_by")
    def handle_blocked_by(cls, value):
        return {str(x) for x in value}

    def dict(self, *args, **kwargs):
        data = super().dict(*args, **kwargs)
        # Keep the config file smaller by skipping values that are defaults.
        for field in (
            "use_multi_node_manager",
            "spark_config",
            "append_job_name",
            "append_output_dir",
            "ext",
        ):
            if data[field] == GenericCommandParametersModel.__fields__[field].default:
                data.pop(field)
        return data



"""Contains logging configuration data."""

import logging
import logging.config

from jade.extensions.registry import Registry


def setup_logging(
    name, filename, console_level=logging.INFO, file_level=logging.INFO, mode="w", packages=None
):
    """Configures logging to file and console.

    Parameters
    ----------
    name : str
        logger name
    filename : str | None
        log filename
    console_level : int, optional
        console log level
    file_level : int, optional
        file log level
    packages : list, optional
        enable logging for these package names

    """
    log_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "short": {
                "format": "%(asctime)s - %(levelname)s [%(name)s "
                "%(filename)s:%(lineno)d] : %(message)s",
            },
            "detailed": {
                "format": "%(asctime)s - %(levelname)s [%(name)s "
                "%(filename)s:%(lineno)d] : %(message)s",
            },
        },
        "handlers": {
            "console": {
                "level": console_level,
                "formatter": "short",
                "class": "logging.StreamHandler",
            },
            "file": {
                "class": "logging.FileHandler",
                "level": file_level,
                "filename": filename,
                "mode": mode,
                "formatter": "detailed",
            },
        },
        "loggers": {
            name: {"handlers": ["console", "file"], "level": "DEBUG", "propagate": False},
        },
    }

    logging_packages = set(Registry().list_loggers())
    if packages is not None:
        for package in packages:
            logging_packages.add(package)

    for package in logging_packages:
        log_config["loggers"][package] = {
            "handlers": ["console", "file"],
            "level": "DEBUG",
            "propagate": False,
        }

    if filename is None:
        log_config["handlers"].pop("file")
        log_config["loggers"][name]["handlers"].remove("file")
        for package in logging_packages:
            if "file" in log_config["loggers"][package]["handlers"]:
                log_config["loggers"][package]["handlers"].remove("file")

    logging.config.dictConfig(log_config)
    logger = logging.getLogger(name)

    return logger


_EVENT_LOGGER_NAME = "_jade_event"


def setup_event_logging(filename, mode="w"):
    """Configures structured event logging.

    Parameters
    ----------
    filename : str
        log filename
    mode : str

    """
    log_config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "basic": {"format": "%(message)s"},
        },
        "handlers": {
            "file": {
                "class": "logging.FileHandler",
                "level": logging.INFO,
                "filename": filename,
                "mode": mode,
                "formatter": "basic",
            },
        },
        "loggers": {
            "_jade_event": {
                "handlers": ["file"],
                "level": "INFO",
                "propagate": False,
            },
        },
    }

    logging.config.dictConfig(log_config)
    logger = logging.getLogger(_EVENT_LOGGER_NAME)

    return logger


def close_event_logging():
    """Close the event log file handle."""
    for handler in logging.getLogger(_EVENT_LOGGER_NAME).handlers:
        handler.close()


def log_event(event):
    """
    Log a structured job event into log file

    Parameters
    ----------
    event: :obj:`StructuredLogEvent`
        An instance of :obj:`StructuredLogEvent`

    """
    logger = logging.getLogger(_EVENT_LOGGER_NAME)
    logger.info(event)



"""Main CLI command for jade."""

import logging
import sys


import click

import jade
from jade.cli.auto_config import auto_config
from jade.cli.cancel_jobs import cancel_jobs
from jade.cli.config import config
from jade.cli.cluster import cluster
from jade.cli.extensions import extensions
from jade.cli.hpc_jobs import hpc_jobs
from jade.cli.pipeline import pipeline
from jade.cli.prune_files import prune_files
from jade.cli.resubmit_jobs import resubmit_jobs
from jade.cli.show_events import show_events
from jade.cli.show_results import show_results
from jade.cli.show_status import show_status
from jade.cli.spark import spark
from jade.cli.db import db
from jade.cli.stats import stats
from jade.cli.submit_jobs import submit_jobs
from jade.cli.try_submit_jobs import try_submit_jobs
from jade.cli.wait import wait


logger = logging.getLogger(__name__)


def _show_version(_, __, version):
    if version:
        print(f"JADE version {jade.__version__}")
        sys.exit(0)
    return version


@click.group()
@click.option(
    "--version",
    callback=_show_version,
    is_flag=True,
    show_default=True,
    help="Show version and exit",
)
def cli(version):
    """JADE commands"""


cli.add_command(extensions)
cli.add_command(auto_config)
cli.add_command(cancel_jobs)
cli.add_command(cluster)
cli.add_command(config)
cli.add_command(db)
cli.add_command(hpc_jobs)
cli.add_command(pipeline)
cli.add_command(prune_files)
cli.add_command(resubmit_jobs)
cli.add_command(show_events)
cli.add_command(show_results)
cli.add_command(show_status)
cli.add_command(spark)
cli.add_command(stats)
cli.add_command(submit_jobs)
cli.add_command(try_submit_jobs)
cli.add_command(wait)



"""Manages PBS jobs."""

import logging

from jade.enums import Status
from jade.hpc.common import HpcJobStatus
from jade.hpc.hpc_manager_interface import HpcManagerInterface
from jade.utils.subprocess_manager import run_command
from jade.utils import utils


logger = logging.getLogger(__name__)

PBS_SCRIPT = """#!/bin/bash
#PBS -N {name} # job name
#PBS -A {alloc} # allocation account
#PBS -l {qos}
#PBS -q {queue} # queue (debug, short, batch, or long)
#PBS -o {path}/{name}_$PBS_JOBID.o
#PBS -e {path}/{name}_$PBS_JOBID.e
{qos_str}
{feature}
echo Running on: $HOSTNAME, Machine Type: $MACHTYPE
{script}

wait
"""


class PbsManager(HpcManagerInterface):
    """Manages PBS jobs."""

    _STATUSES = {
        "Q": HpcJobStatus.QUEUED,
        "R": HpcJobStatus.RUNNING,
        "C": HpcJobStatus.COMPLETE,
    }

    def __init__(self, config):
        self._config = config

    def am_i_manager(self):
        assert False

    def cancel_job(self, job_id):
        return 0

    def check_status(self, name=None, job_id=None):
        qstat_rows = self._qstat()
        if qstat_rows is None:
            return HpcJobStatus.NONE

        # TODO job_id

        return self._get_status_from_output(qstat_rows, name)

    def check_statuses(self):
        assert False

    def check_storage_configuration(self):
        pass

    def create_cluster(self):
        pass

    def create_local_cluster(self):
        pass

    def get_config(self):
        return self._config

    def get_current_job_id(self):
        assert False

    def get_job_stats(self, job_id):
        return None

    def get_local_scratch(self):
        return "."

    @staticmethod
    def get_num_cpus():
        return 18

    @staticmethod
    def _get_status_from_output(qstat_rows, name):
        # column location of various job identifiers
        col_loc = {"id": 0, "name": 3}

        # reverse the list so most recent jobs are first
        qstat_rows.reverse()

        # update job status from qstat list
        status = HpcJobStatus.UNKNOWN
        for row in qstat_rows:
            row = row.split()
            # make sure the row is long enough to be a job status listing
            # TODO regex?
            if len(row) > 10:
                if row[col_loc["name"]].strip() == name.strip():
                    # Job status is located at the -2 index
                    status = PbsManager._STATUSES.get(row[-2], HpcJobStatus.UNKNOWN)
                    if status is HpcJobStatus.UNKNOWN:
                        logger.error("Unknown PBS job status: %s", row[-2])
                    break
        return status

    def _qstat(self):
        """Run the PBS qstat command and return the stdout split to rows.

        Returns
        -------
        qstat_rows : list | None
            List of strings where each string is a row in the qstat printout.
            Returns None if qstat is empty.

        """
        cmd = "qstat -u {user}".format(user=self.USER)
        output = {}
        run_command(cmd, output)
        if not output["stdout"]:
            # No jobs are currently running.
            return None

        qstat_rows = output["stdout"].split("\n")
        return qstat_rows

    def create_submission_script(self, name, script, filename, path="."):
        feature = self._config.get("feature")
        if feature is None:
            feature = ""
        else:
            feature = "#PBS -l feature={}".format("feature")

        script = PBS_SCRIPT.format(
            name=name,
            alloc=self._config["allocation"],
            qos=self._config["qos"],
            queue=self._config["queue"],
            path=path,
            feature=feature,
            script=script,
        )

        utils.create_script(filename, script)

    def list_active_nodes(self, job_id):
        assert False

    def log_environment_variables(self):
        pass

    def submit(self, filename):
        output = {}
        ret = run_command("qsub {}".format(filename), output)
        if ret == 0:
            result = Status.GOOD
            job_id = output["stdout"]
        else:
            result = Status.ERROR
            job_id = None

        return result, job_id, output["stderr"]



#!/usr/bin/env python

"""Creates the JADE configuration for stage 2 of the demo pipeline."""

import os
import sys

from jade.models import PipelineConfig
from jade.utils.subprocess_manager import run_command
from jade.utils.utils import load_data

PRED_GDP_COMMANDS_FILE = "pred_gdp_commands.txt"


def main():
    config = PipelineConfig(**load_data(os.environ["JADE_PIPELINE_STATUS_FILE"]))
    cur_stage = config.stages[-1]
    cur_stage_output = cur_stage.path
    previous_stage = config.stages[-2]
    previous_stage_output = previous_stage.path
    script = "jade/extensions/demo/merge_pred_gdp.py"

    with open(PRED_GDP_COMMANDS_FILE, "w") as f_out:
        cmd = f"python {script} run {previous_stage_output} {cur_stage_output}"
        f_out.write(cmd + "\n")

    cmd = "jade config create pred_gdp_commands.txt -c config-stage2.json"
    sys.exit(run_command(cmd))


if __name__ == "__main__":
    main()



"""CLI to automatically create a configuration."""

import logging

import click

from jade.common import CONFIG_FILE
from jade.loggers import setup_logging
from jade.exceptions import InvalidExtension
from jade.extensions.registry import Registry, ExtensionClassType
from jade.jobs.job_post_process import JobPostProcess


# TODO: need one group command for auto-config; this should be a subcommand.


@click.command()
@click.argument("extension")
@click.argument("inputs", nargs=-1)
@click.option(
    "-p",
    "--job-post-process-config-file",
    type=click.Path(exists=True),
    is_eager=True,
    help="The path of job-based post-process config file.",
)
@click.option(
    "-c",
    "--config-file",
    default=CONFIG_FILE,
    show_default=True,
    help="config file to generate.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable verbose log output.",
)
def auto_config(extension, inputs, job_post_process_config_file, config_file, verbose):
    """Automatically create a configuration."""
    level = logging.DEBUG if verbose else logging.WARNING
    setup_logging("auto_config", None, console_level=level)

    if job_post_process_config_file is not None:
        module, class_name, data = JobPostProcess.load_config_from_file(
            job_post_process_config_file
        )
        JobPostProcess(module, class_name, data)  # ensure everything ok
        job_post_process_config = {"module": module, "class": class_name, "data": data}
    else:
        job_post_process_config = None

    # User extension
    registry = Registry()
    if not registry.is_registered(extension):
        raise InvalidExtension(f"Extension '{extension}' is not registered.")

    cli = registry.get_extension_class(extension, ExtensionClassType.CLI)
    config = cli.auto_config(*inputs, job_post_process_config=job_post_process_config)
    print(f"Created configuration with {config.get_num_jobs()} jobs.")
    config.dump(config_file)
    print(f"Dumped configuration to {config_file}.\n")



"""Contains class for running any post process scripts available"""

import os
import logging

from prettytable import PrettyTable

from jade.common import JOBS_OUTPUT_DIR, OUTPUT_DIR
from jade.utils.utils import load_data, output_to_file

logger = logging.getLogger(__name__)


class JobPostProcess:
    """Class used to dynamically run post process scripts"""

    _results_file = "post-process-results.json"

    def __init__(
        self, module_name, class_name, data=None, output=OUTPUT_DIR, job_name=None, *args, **kwargs
    ):
        """Constructs JobPostProcess

        Parameters
        ----------
        module_name : str
            module which contains class that needs to run
        class_name : str
            class implementing post process to run
        data : dict
            optional dictionary of overrides data to send to post process

        """
        self._data = data or {}
        self._job_name = job_name
        self._output = output

        try:
            process_module = __import__(module_name, fromlist=[class_name])
            process_class = getattr(process_module, class_name)
            self._post_process = process_class(overrides=self._data, job_name=self._job_name)
        except ModuleNotFoundError:
            logger.exception("Could not import module %s", module_name)
            raise
        except ValueError:
            logger.exception("Module %s does not have class %s", module_name, class_name)
            raise

    def run(self, *args, **kwargs):
        """Runs post-process class' run function"""
        self._post_process.run(*args, **kwargs)
        self._dump_results()

    def serialize(self):
        """Create data for serialization."""
        serialized_data = {
            "class": self._post_process.__class__.__name__,
            "module": self._post_process.__module__,
        }

        data = self._post_process.serialized_data
        if data is not None:
            serialized_data["data"] = data

        return serialized_data

    def dump_config(self, output_file="post-process-config.toml"):
        """Outputs post process data to results file

        Parameters
        ----------
        output_file : str
        """
        output_to_file(self.serialize(), output_file)

    def _get_job_results_dir(self):
        return os.path.join(self._output, self._job_name)

    def _dump_results(self):
        results = self._post_process.get_results()
        output_path = os.path.join(self._get_job_results_dir(), self._results_file)
        data = {
            "job": self._job_name,
            "post-process": type(self._post_process).__name__,
            "results": results,
        }
        logger.info("Dumping post-process results to %s", output_path)
        output_to_file(data, output_path)

    @classmethod
    def load_config_from_file(cls, config_file):
        """Loads config from given toml file

        Parameters
        ----------
        config_file : str

        Returns
        -------
        module_name : str
            module which contains class that needs to run
        class_name : str
            class implementing post process to run
        data : dict
            optional dictionary of additional data to send to post process

        """

        config = load_data(config_file)
        module_name = None
        class_name = None
        data = {}

        if "module" in config.keys():
            module_name = config["module"]

        if "class" in config.keys():
            class_name = config["class"]

        if "data" in config.keys():
            for data_index in config["data"]:
                data[data_index] = config["data"][data_index]

        return module_name, class_name, data

    @classmethod
    def show_results(cls, output_dir, job_name=None, input_file=None):
        """Show the post process results for jobs in a table.
           Expects that JOBS_OUTPUT_DIR contains folders named after
           jobs containing results.

        Parameters
        ----------
        job_name : str
            optional individual job to display
        input_file : str
            optional input file name
        """

        if input_file is None:
            input_file = cls._results_file

        job_results_dir = os.path.join(output_dir, JOBS_OUTPUT_DIR)
        print(f"Post-process results from directory: {job_results_dir}")
        job_names = None

        for _, dirs, _ in os.walk(job_results_dir):
            job_names = dirs
            assert job_names
            break

        table = PrettyTable()

        for job in job_names:
            if job_name is not None and job_name != job:
                continue

            post_process_results = load_data(os.path.join(job_results_dir, job, input_file))
            results = post_process_results["results"]["outputs"]
            if not table.field_names and results:
                table.field_names = ["job"] + list(results[0].keys())

            for result in results:
                row = [job]
                for column in result:
                    row.append(result[column])

                table.add_row(row)

        print(table)



"""CLI to display and manage config files."""

import json
import logging
import os
import sys
import tempfile
from pathlib import Path

import click
from prettytable import PrettyTable

from jade.cli.common import COMMON_SUBMITTER_OPTIONS, add_options, make_submitter_params
from jade.jobs.cluster import Cluster
from jade.common import CONFIG_FILE
from jade.extensions.generic_command import GenericCommandConfiguration, GenericCommandParameters
from jade.jobs.job_configuration_factory import create_config_from_file
from jade.loggers import setup_logging
from jade.models import HpcConfig, SlurmConfig, FakeHpcConfig, LocalHpcConfig
from jade.models.submission_group import SubmissionGroup
from jade.utils.utils import dump_data, load_data


logger = logging.getLogger(__name__)


def _handle_indexes_list(_, __, indexes):
    if indexes is None:
        return indexes
    return [int(x) for x in indexes]


def _handle_indexes_set(_, __, indexes):
    if indexes is None:
        return indexes
    return {int(x) for x in indexes}


@click.group()
def config():
    """Manage a JADE configuration."""
    setup_logging("config", None)


@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "-A",
    "--append-job-name",
    default=False,
    is_flag=True,
    show_default=True,
    help="Set append_job_name for every job.",
)
@click.option(
    "-a",
    "--append-output-dir",
    default=False,
    is_flag=True,
    show_default=True,
    help="Set append_output_dir for every job.",
)
@click.option(
    "-c",
    "--config-file",
    default=CONFIG_FILE,
    show_default=True,
    help="config file to generate.",
)
@click.option(
    "-C",
    "--cancel-on-blocking-job-failure",
    is_flag=True,
    default=False,
    show_default=True,
    help="cancel any job if one of its blocking jobs fails.",
)
@click.option(
    "-m",
    "--minutes-per-job",
    type=int,
    help="estimated minutes per job.",
)
@click.option(
    "--shuffle/--no-shuffle",
    is_flag=True,
    default=False,
    show_default=True,
    help="Shuffle order of jobs.",
)
@click.option(
    "--strip-whitespace/--no-strip-whitespace",
    is_flag=True,
    default=False,
    show_default=True,
    help="Strip whitespace in file.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable verbose log output.",
)
def create(
    filename,
    append_job_name,
    append_output_dir,
    config_file,
    cancel_on_blocking_job_failure,
    minutes_per_job,
    shuffle,
    strip_whitespace,
    verbose,
):
    """Create a config file from a filename with a list of executable commands."""
    level = logging.DEBUG if verbose else logging.WARNING
    setup_logging("auto_config", None, console_level=level)

    config = GenericCommandConfiguration.auto_config(
        filename,
        cancel_on_blocking_job_failure=cancel_on_blocking_job_failure,
        minutes_per_job=minutes_per_job,
        append_job_name=append_job_name,
        append_output_dir=append_output_dir,
    )
    if shuffle:
        config.shuffle_jobs()
    print(f"Created configuration with {config.get_num_jobs()} jobs.")
    indent = None if strip_whitespace else 2
    config.dump(config_file, indent=indent)
    print(f"Dumped configuration to {config_file}.\n")


@click.command()
@click.argument("config_file", type=click.Path(exists=True))
@click.argument("job_index", type=int)
@click.argument("blocking_job_indexes", nargs=-1, callback=_handle_indexes_set)
@click.option(
    "-o",
    "--output-file",
    default="updated_config.json",
    show_default=True,
    help="Create new config file with updated jobs.",
)
def assign_blocked_by(config_file, job_index, blocking_job_indexes, output_file):
    """Assign the blocked_by attribute for job_name with jobs specified by block_job_indexes
    (0-based). If no blocking jobs are specified then make all other jobs blocking.

    \b
    Examples:
    1. Make the last job blocked by all other jobs.
       jade config assign-blocked-by config.json post-process-job -o new-config.json
    2. Select first 10 indexes through shell expansion.
       jade config assign-blocked-by config.json post-process-job {0..9} -o new-config.json
    3. Specify blocking indexes.
       jade config assign-blocked-by config.json post-process-job 0 1 2 3 -o new-config.json
    """
    if blocking_job_indexes and job_index in blocking_job_indexes:
        print(f"job_index={job_index} is included in blocking_job_indexes", file=sys.stderr)
        sys.exit(1)

    config = create_config_from_file(config_file)
    if job_index > config.get_num_jobs() - 1:
        print(f"Invalid job_index={job_index}. Max={config.get_num_jobs() - 1}", file=sys.stderr)
        sys.exit(1)

    blocking_jobs = set()
    main_job = None
    for i, job in enumerate(config.iter_jobs()):
        if i == job_index:
            main_job = job
            continue
        if i != job_index and (not blocking_job_indexes or i in blocking_job_indexes):
            blocking_jobs.add(job.name)

    assert main_job
    main_job.set_blocking_jobs(blocking_jobs)
    config.dump(output_file, indent=2)
    print(f"Added {len(blocking_jobs)} blocking jobs to {main_job.name} in {output_file}")


@click.command()
@click.option(
    "-a",
    "--account",
    default="",
    help="HPC account/allocation",
)
@click.option(
    "-c",
    "--config-file",
    default="hpc_config.toml",
    show_default=True,
    help="config file to create",
)
@click.option(
    "-m",
    "--mem",
    default=None,
    help="Amount of memory required by a single node.",
)
@click.option(
    "-p",
    "--partition",
    default=None,
    help="HPC partition",
)
@click.option(
    "-q",
    "--qos",
    default=None,
    type=str,
    help="QoS value",
)
@click.option(
    "-t",
    "--hpc-type",
    type=click.Choice(["slurm", "fake", "local"]),
    default="slurm",
    show_default=True,
    help="HPC queueing system",
)
@click.option(
    "--tmp",
    default=None,
    help="Amount of local storage space required by a single node.",
)
@click.option(
    "-w",
    "--walltime",
    default="4:00:00",
    help="HPC walltime",
)
def hpc(account, config_file, mem, partition, qos, hpc_type, tmp, walltime):
    """Create an HPC config file."""
    if hpc_type == "slurm":
        hpc = SlurmConfig(
            account=account,
            mem=mem,
            partition=partition,
            qos=qos,
            tmp=tmp,
            walltime=walltime,
        )
    elif hpc_type == "fake":
        hpc = FakeHpcConfig(walltime=walltime)
    else:
        assert hpc_type == "local"
        hpc = LocalHpcConfig()

    # This converts enums to values.
    data = json.loads(HpcConfig(hpc_type=hpc_type, hpc=hpc).json())
    dump_data(data, config_file)
    print(f"Created HPC config file {config_file}")


@click.command()
@click.argument("config_file", type=click.Path(exists=True))
@click.option(
    "-f",
    "--fields",
    multiple=True,
    help="include in output table; can specify mulitple times",
)
@click.option(
    "-n",
    "--no-blocked-by",
    is_flag=True,
    default=False,
    show_default=True,
    help="exclude blocking jobs in table",
)
def show(config_file, fields, no_blocked_by):
    """Show the jobs in the configuration."""
    _show(config_file, fields, not no_blocked_by)


# This is a standalone function so that it can be called from _filter.
def _show(config_file, fields=None, blocked_by=True):
    config = create_config_from_file(config_file)
    num_jobs = config.get_num_jobs()
    print(f"Num jobs: {num_jobs}")
    if num_jobs == 0:
        return

    if config.setup_command is not None:
        print(f"Setup command: {config.setup_command}")
    if config.teardown_command is not None:
        print(f"Teardown command: {config.teardown_command}")
    print()

    # generic_command jobs have a command field which is very useful.
    # Other extensions do not.
    has_command = False
    for job in config.iter_jobs():
        if isinstance(job, GenericCommandParameters):
            has_command = True
            break

    field_names = ["index", "name"]
    if has_command:
        field_names.append("command")
    if blocked_by:
        field_names.append("blocked_by (job names)")
    if fields is not None:
        field_names += fields

    table = PrettyTable()
    table.field_names = field_names
    for i, job in enumerate(config.iter_jobs()):
        job_dict = job.serialize()
        row = [i, job.name]
        if has_command:
            row.append(job_dict.get("command", ""))
        if blocked_by:
            blocking_jobs = sorted(list(job.get_blocking_jobs()))
            text = ", ".join(blocking_jobs)
            if len(text) > 50:
                text = f"truncated...blocked by {len(blocking_jobs)} jobs"
            row.append(text)
        if fields is not None:
            for field in fields:
                row.append(job_dict.get(field, ""))
        table.add_row(row)
    print(table)


@click.command("filter")
@click.argument("config_file", type=click.Path(exists=True))
@click.argument("indexes", nargs=-1, callback=_handle_indexes_list)
@click.option(
    "-o",
    "--output-file",
    help="Create new config file with filtered jobs.",
)
@click.option(
    "-f",
    "--fields",
    type=str,
    multiple=True,
    nargs=2,
    help="Filter on field value. Multiple accepted.",
)
@click.option(
    "-s",
    "--show-config",
    is_flag=True,
    show_default=True,
    default=False,
    help="Show the new config (only applicable if output-file is provided).",
)
# Named _filter to avoid collisions with the built-in function.
def _filter(config_file, output_file, indexes, fields, show_config=False):
    """Filters jobs in CONFIG_FILE. Prints the new jobs to the console or
    optionally creates a new file.

    Note: This does not detect duplicate ranges.

    \b
    Examples:
    1. Select the first two jobs by index. Output only.
       jade config filter c1.json 0 1
    2. Select indexes through shell expansion, create new file.
       jade config filter c1.json {0..4} {10..14} 20 25 -o c2.json
    3. Select jobs with parameters param1=green and param2=3.
       jade config filter c1.json -o c2.json -f param1 green -f param2 3

    """
    cfg = load_data(config_file)
    jobs = cfg["jobs"]
    if not jobs:
        print("The configuration has no jobs", file=sys.stderr)
        sys.exit(1)

    if output_file is None:
        handle, new_config_file = tempfile.mkstemp(suffix=".json")
        os.close(handle)
        show_config = True
    else:
        new_config_file = output_file

    try:
        if not new_config_file.endswith(".json"):
            print("new_config_file must have extension .json", file=sys.stderr)
            sys.exit(1)

        orig_len = len(jobs)

        # Note: when looking at just the JSON, there is no way to get the job name,
        # and so we can't check for duplicates.

        if indexes:
            new_jobs = [jobs[i] for i in indexes]
        else:
            new_jobs = jobs

        if fields:
            final_jobs = []
            for job in new_jobs:
                matched = True
                for field in fields:
                    if str(job[field[0]]) != field[1]:
                        matched = False
                        break
                if matched:
                    final_jobs.append(job)

            new_jobs = final_jobs

        cfg["jobs"] = new_jobs
        new_len = len(cfg["jobs"])
        dump_data(cfg, new_config_file, indent=4)
        print(f"Filtered {config_file} ({orig_len} jobs) into ({new_len} jobs)\n")
        if output_file is not None:
            print(f"Wrote new config to {output_file}")

        if show_config:
            _show(new_config_file)
    finally:
        if output_file is None:
            os.remove(new_config_file)


@click.command()
@click.option(
    "-c",
    "--config-file",
    default="submitter_params.json",
    show_default=True,
    type=Path,
    help="config file to create; can be .toml or .json",
)
@add_options(COMMON_SUBMITTER_OPTIONS)
def submitter_params(
    config_file=None,
    dry_run=None,
    per_node_batch_size=None,
    hpc_config=None,
    local=None,
    max_nodes=None,
    poll_interval=None,
    resource_monitor_interval=None,
    resource_monitor_type=None,
    resource_monitor_stats=None,
    num_parallel_processes_per_node=None,
    verbose=None,
    reports=None,
    enable_singularity=None,
    container=None,
    try_add_blocked_jobs=None,
    time_based_batching=None,
    node_setup_script=None,
    node_shutdown_script=None,
    no_distributed_submitter=None,
):
    """Create parameters for use in 'jade submit-jobs'."""
    params = make_submitter_params(
        per_node_batch_size=per_node_batch_size,
        dry_run=dry_run,
        hpc_config=hpc_config,
        local=local,
        max_nodes=max_nodes,
        poll_interval=poll_interval,
        resource_monitor_interval=resource_monitor_interval,
        resource_monitor_type=resource_monitor_type,
        resource_monitor_stats=resource_monitor_stats,
        num_parallel_processes_per_node=num_parallel_processes_per_node,
        verbose=verbose,
        reports=reports,
        enable_singularity=enable_singularity,
        container=container,
        try_add_blocked_jobs=try_add_blocked_jobs,
        time_based_batching=time_based_batching,
        node_setup_script=node_setup_script,
        node_shutdown_script=node_shutdown_script,
        no_distributed_submitter=no_distributed_submitter,
    )
    # This converts enums to values.
    data = json.loads(params.json())
    if config_file.suffix == ".json":
        dump_data(data, config_file, indent=2)
    else:
        dump_data(data, config_file)
    print(f"Created submitter parameter file {config_file}")


@click.command()
@click.argument("output_dir", type=click.Path(exists=True), callback=lambda _, __, x: Path(x))
@click.option(
    "-c",
    "--config-file",
    default="submission_groups.json",
    show_default=True,
    type=Path,
    help="config file to create; can be .toml or .json",
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    show_default=True,
    help="Overwrite config_file if it already exists.",
)
def save_submission_groups(output_dir, config_file, force):
    if config_file.exists() and not force:
        print(
            f"{config_file} exists. Use a different name or pass --force to overwrite.",
            file=sys.stderr,
        )
        sys.exit(1)

    existing_groups_file = output_dir / Cluster.SUBMITTER_GROUP_FILE
    if not existing_groups_file.exists():
        print(f"{output_dir} is not a valid JADE output directory", file=sys.stderr)
        sys.exit(1)

    data = load_data(existing_groups_file)
    dump_data(data, config_file, indent=2)
    print(f"Copied submission groups to {config_file}")


@click.command()
@click.argument("params_file", type=click.Path(exists=True))
@click.argument("name")
@click.argument("config_file", type=click.Path(exists=True))
def add_submission_group(params_file, name, config_file):
    """Add a submission group with parameters defined in params_file to config_file."""
    config = load_data(config_file)
    for group in config["submission_groups"]:
        if name == group["name"]:
            print(f"Error: {name} is already stored in {config_file}", file=sys.stderr)
            sys.exit(1)

    params = load_data(params_file)
    group = {
        "name": name,
        "submitter_params": params,
    }
    # Make sure it parses.
    SubmissionGroup(**group)

    config["submission_groups"].append(group)
    dump_data(config, config_file, indent=2)
    print(f"Updated {config_file} with submission group {name}.")


config.add_command(create)
config.add_command(assign_blocked_by)
config.add_command(hpc)
config.add_command(show)
config.add_command(_filter)
config.add_command(submitter_params)
config.add_command(save_submission_groups)
config.add_command(add_submission_group)



"""
Implement JobConfiguration for auto-regression analysis.
"""
from jade.jobs.job_container_by_key import JobContainerByKey
from jade.jobs.job_configuration import JobConfiguration
from jade.extensions.demo.autoregression_parameters import AutoRegressionParameters


class AutoRegressionConfiguration(JobConfiguration):
    """
    A class used to configure auto-regression jobs
    """

    def _serialize(self, data):
        pass

    def create_from_result(self, job, output_dir):
        return None



"""Manages the execution of a pipeline of JADE configurations."""

from jade.jobs.job_configuration_factory import create_config_from_file
import logging
import os
import shutil

from jade.exceptions import ExecutionError, InvalidParameter
from jade.jobs.job_submitter import JobSubmitter
from jade.models.pipeline import PipelineStage, PipelineConfig
from jade.models.submitter_params import SubmitterParams
from jade.result import Result, serialize_result
from jade.utils.subprocess_manager import run_command
from jade.utils.timing_utils import timed_info
from jade.utils.utils import dump_data, load_data


logger = logging.getLogger(__name__)


class PipelineManager:
    """Manages the execution of a pipeline of JADE configurations."""

    CONFIG_FILENAME = "pipeline.json"

    def __init__(self, config_file, output):
        self._output = output
        self._config_file = config_file
        self._config = self._deserialize()
        self._config.path = output

    @classmethod
    def create(cls, config_file, output):
        """Create a new PipelineManager.

        Parameters
        ----------
        config_file : str
        output : str
            output directory for execution

        Returns
        -------
        PipelineManager

        """
        os.makedirs(output, exist_ok=True)
        main_file = os.path.join(output, cls.CONFIG_FILENAME)
        shutil.copyfile(config_file, main_file)
        mgr = cls(main_file, output)
        for stage in mgr.stages:
            stage.path = cls.get_stage_output_path(output, stage.stage_num)
        mgr._serialize()
        return mgr

    @classmethod
    def load(cls, output):
        """Load a PipelineManager from an execution directory.

        Parameters
        ----------
        output : str

        Returns
        -------
        PipelineManager

        """
        config_file = os.path.join(output, cls.CONFIG_FILENAME)
        return cls(config_file, output)

    @property
    def config(self):
        return self._config

    def submit_next_stage(self, stage_num, return_code=None):
        """Submit the next stage of the pipeline for execution.

        Parameters
        ----------
        stage_num : int
            stage number to submit
        return_code : int
            status of the previous stage if this wasn't the first

        """
        # There is a challenge in passing the information that each possible
        # stage/extension might require. This solution attempts to be as
        # flexible as possible without requiring a specific interface.
        #
        # Set these environment variables so that the auto-config scripts
        # can extract information they need to create execution script
        # arguments.
        os.environ["JADE_PIPELINE_OUTPUT_DIR"] = self._output
        os.environ["JADE_PIPELINE_STATUS_FILE"] = self._config_file
        try:
            self._submit_next_stage(stage_num, return_code=return_code)
        finally:
            os.environ.pop("JADE_PIPELINE_OUTPUT_DIR")
            os.environ.pop("JADE_PIPELINE_STATUS_FILE")
            if "JADE_PIPELINE_STAGE_ID" in os.environ:
                os.environ.pop("JADE_PIPELINE_STAGE_ID")

    @staticmethod
    def create_config_from_files(config_files, pipeline_config_file, submit_params):
        """Create a new PipelineConfig from a list of config files.

        Parameters
        ----------
        config_files : list
            Jade config files for each stage of the pipeline
        pipeline_config_file : str
        submit_params : SubmitterParams

        Returns
        -------
        PipelineConfig

        """
        stages = []
        for i, config_file in enumerate(config_files):
            stage_num = i + 1
            stages.append(
                PipelineStage(
                    config_file=config_file,
                    stage_num=stage_num,
                    submitter_params=submit_params,
                    auto_config_cmd=None,
                )
            )

        config = PipelineConfig(stages=stages, stage_num=1)
        with open(pipeline_config_file, "w") as f_out:
            f_out.write(config.json(indent=2))
        logger.info("Created pipeline config file %s", pipeline_config_file)

    @staticmethod
    def create_config_from_commands(auto_config_cmds, pipeline_config_file, submit_params):
        """Create a new PipelineConfig from a list of jade config commands.

        Parameters
        ----------
        auto_config_cmds : list
            list of commands (str) used to create Jade configs.
        pipeline_config_file : str
        submit_params : SubmitterParams

        Returns
        -------
        PipelineConfig

        """
        stages = []
        for i, cmd in enumerate(auto_config_cmds):
            stage_num = i + 1
            stages.append(
                PipelineStage(
                    auto_config_cmd=cmd,
                    config_file=PipelineManager.get_stage_config_file_name(stage_num),
                    stage_num=stage_num,
                    submitter_params=submit_params,
                )
            )

        config = PipelineConfig(stages=stages, stage_num=1)
        with open(pipeline_config_file, "w") as f_out:
            f_out.write(config.json(indent=2))
        logger.info("Created pipeline config file %s", pipeline_config_file)

    def _deserialize(self):
        return PipelineConfig(**load_data(self._config_file))

    def _serialize(self):
        print(self.stage_num)
        with open(self._config_file, "w") as f_out:
            f_out.write(self._config.json(indent=2))

    def _submit_next_stage(self, stage_num, return_code=None):
        if return_code is None:
            assert stage_num == 1, str(stage_num)
        else:
            if stage_num != self.stage_num + 1:
                raise InvalidParameter(
                    f"expected stage_num {self.stage_num + 1}, received {stage_num}"
                )

            self._config.stages[stage_num - 2].return_code = return_code
            self._config.stage_num += 1

        if self._config.stage_num == len(self._config.stages) + 1:
            logger.info("Pipeline is complete")
            self._config.is_complete = True
            self._serialize()
            return

        logger.info("Start execution pipeline stage %s/%s", stage_num, len(self._config.stages))

        self._serialize()
        stage = self._config.stages[self.stage_num - 1]
        os.environ["JADE_PIPELINE_STAGE_ID"] = str(self.stage_num)
        if stage.auto_config_cmd is not None:
            self._run_auto_config(stage)
        output = self.get_stage_output_path(self.path, self.stage_num)
        config = create_config_from_file(stage.config_file)
        if not config.submission_groups:
            config.assign_default_submission_group(stage.submitter_params)
        ret = JobSubmitter.run_submit_jobs(
            config,
            output,
            pipeline_stage_num=self.stage_num,
        )
        if ret != 0:
            raise ExecutionError(f"stage {self.stage_num} failed")

    def _run_auto_config(self, stage):
        if os.path.exists(stage.config_file):
            os.remove(stage.config_file)

        ret = run_command(stage.auto_config_cmd)
        if ret != 0:
            raise ExecutionError(f"Failed to auto-config stage {self.stage_num}: {ret}")

        if not os.path.exists(stage.config_file):
            raise ExecutionError(
                f"auto-config stage {self.stage_num} did not produce {stage.config_file}"
            )

        final_file = self.get_stage_config_file_path(self._output, self.stage_num)
        shutil.copyfile(stage.config_file, final_file)
        stage.config_file = final_file

    @property
    def config(self):
        """Return the pipeline config.

        Returns
        -------
        PipelineConfig

        """
        return self._config

    @property
    def path(self):
        """Return the pipeline directory.

        Returns
        -------
        int

        """
        return self._config.path

    @property
    def stage_num(self):
        """Return the current stage index

        Returns
        -------
        int

        """
        return self._config.stage_num

    @property
    def stages(self):
        """Return the stages in the pipeline.

        Returns
        -------
        list
            list of PipelineStage

        """
        return self._config.stages

    @staticmethod
    def get_stage_config_file_name(stage_num):
        """Return the filename of a stage config file."""
        return f"config-stage{stage_num}.json"

    @staticmethod
    def get_stage_config_file_path(output, stage_num):
        """Return the path to a stage config file."""
        return os.path.join(output, PipelineManager.get_stage_config_file_name(stage_num))

    @staticmethod
    def get_stage_output_name(stage_num):
        """Return the output directory name of a stage."""
        return f"output-stage{stage_num}"

    @staticmethod
    def get_stage_output_path(output, stage_num):
        """Return the path to the output directory of a stage."""
        return os.path.join(output, PipelineManager.get_stage_output_name(stage_num))



"""Defines interface for Configuration Inputs."""

import abc


class JobInputsInterface(abc.ABC):
    """Interface definition for configuration inputs."""

    @property
    def base_directory(self):
        """Return the base directory."""

    @abc.abstractmethod
    def get_available_parameters(self):
        """Return a dictionary containing all available parameters."""



"""CLI to wait for completion of jobs."""

import sys
import time

import click

from jade.common import OUTPUT_DIR
from jade.exceptions import InvalidConfiguration
from jade.jobs.cluster import Cluster


@click.command()
@click.option(
    "-o",
    "--output",
    default=OUTPUT_DIR,
    show_default=True,
    help="directory containing submission output",
)
@click.option(
    "-p",
    "--poll-interval",
    default=10.0,
    show_default=True,
    help="poll interval in minutes",
)
@click.option(
    "-t",
    "--timeout",
    type=int,
    help="Timeout in minutes. Default is no timeout.",
)
def wait(output, poll_interval, timeout):
    """Wait for a JADE submission to complete."""
    if timeout is not None:
        end = time.time() + timeout * 60
    else:
        end = None
    while True:
        try:
            cluster, _ = Cluster.deserialize(output)
        except InvalidConfiguration:
            print(f"{output} is not a JADE output directory used in cluster mode", file=sys.stderr)
            sys.exit(1)
        if cluster.is_complete():
            print("All jobs are complete")
            break
        time.sleep(poll_interval * 60)
        if end is not None and time.time() > end:
            sys.exit(1)



"""Defines class for managing a job queue."""

from collections import OrderedDict
import logging
import os
import time

from jade.enums import Status

logger = logging.getLogger(__name__)


DEFAULT_POLL_INTERVAL = 10
DEFAULT_MONITOR_INTERVAL = 10


class JobQueue:
    """Submits jobs for execution in parallel.

    There are two ways to use this class:

    1. Build a list of jobs and pass that list to :meth:`JobQueue.run_jobs`.
       It will run to completion.
    2. Call :meth:`JobQueue.submit` as jobs become ready to run. JobQueue will
       either run it immediately or queue it if too many commands are
       outstanding.  In this mode it is up to the caller to call
       :meth:`JobQueue.process_queue` periodically. That will look for job
       completions pull new jobs off the queue. JobQueue does not start a
       background thread to do this automatically.

    """

    def __init__(
        self,
        max_queue_depth,
        existing_jobs=None,
        poll_interval=DEFAULT_POLL_INTERVAL,
        monitor_func=None,
        monitor_interval=DEFAULT_MONITOR_INTERVAL,
    ):
        """
        Parameters
        ----------
        max_queue_depth : int
            Maximum number of sub-processes to maintain
        poll_interval : int
            Seconds to sleep in between completion checks.
        monitor_interval : int
            Seconds to sleep in between resource monitor cycles.
            Can be overridden with environment variable JADE_MONITOR_INTERVAL.

        """
        self._queue_depth = max_queue_depth
        self._poll_interval = poll_interval
        self._outstanding_jobs = OrderedDict()
        self._queued_jobs = []
        self._num_jobs = 0
        self._num_completed = 0
        self._monitor_func = monitor_func
        self._last_monitor_time = None

        if existing_jobs is not None:
            for job in existing_jobs:
                self._outstanding_jobs[job.name] = job

        interval = os.environ.get("JADE_MONITOR_INTERVAL")
        if interval:
            self._monitor_interval = int(interval)
        else:
            self._monitor_interval = monitor_interval

        if self._monitor_interval is not None:
            self._poll_interval = min(self._poll_interval, self._monitor_interval)
        logger.debug(
            "queue_depth=%s poll_interval=%s monitor_interval=%s",
            self._queue_depth,
            self._poll_interval,
            self._monitor_interval,
        )

    def _check_completions(self):
        logger.debug("check for completions")
        failed_jobs = set()
        # If jobs fail and are configured to cancel blocked jobs, we may need to run this
        # loop many times to cancel the entire chain.
        need_to_rerun = True
        while need_to_rerun:
            need_to_rerun = False
            completed_jobs = []
            for name, job in self._outstanding_jobs.items():
                if job.is_complete():
                    completed_jobs.append(name)
                    if job.return_code != 0:
                        failed_jobs.add(job.name)

            self._num_completed += len(completed_jobs)
            logger.debug("found num_completed=%s", len(completed_jobs))
            for name in completed_jobs:
                self._outstanding_jobs.pop(name)
                logger.debug("Completed a job %s", name)

                canceled_indices = []
                for i, job in enumerate(self._queued_jobs):
                    blocking_jobs = job.get_blocking_jobs()
                    if blocking_jobs:
                        if job.cancel_on_blocking_job_failure and blocking_jobs.intersection(
                            failed_jobs
                        ):
                            job.set_blocking_jobs(set())
                            job.cancel()
                            canceled_indices.append(i)
                            self._num_jobs += 1
                            self._outstanding_jobs[job.name] = job
                            need_to_rerun = True
                        elif name in blocking_jobs:
                            logger.debug("Remove %s from job=%s blocked list", name, job.name)
                            job.remove_blocking_job(name)

                for index in reversed(canceled_indices):
                    self._queued_jobs.pop(index)

    def _run_job(self, job):
        logger.debug("Run job %s", job.name)
        if job.run() == Status.GOOD:
            self._num_jobs += 1
            self._outstanding_jobs[job.name] = job

    def is_full(self):
        """Return True if the max number of jobs is outstanding.

        Returns
        -------
        bool

        """
        return len(self._outstanding_jobs) >= self._queue_depth

    def _handle_monitor_func(self, force=False):
        if self._monitor_func is None or self._monitor_interval is None:
            return

        needs_monitor = False
        cur_time = time.time()
        if force:
            needs_monitor = True
        elif self._last_monitor_time is None:
            needs_monitor = True
        elif cur_time - self._last_monitor_time > self._monitor_interval:
            needs_monitor = True

        if needs_monitor:
            logger.debug("Run monitor function")
            ids = {x.name: x.get_id() for x in self.outstanding_jobs}
            self._monitor_func(ids=ids)
            self._last_monitor_time = cur_time

    @property
    def outstanding_jobs(self):
        """Return the outstanding jobs.

        Returns
        -------
        dict_values

        """
        return self._outstanding_jobs.values()

    def process_queue(self):
        """Process completions and submit new jobs if the queue is not full."""
        self._check_completions()
        self._handle_monitor_func()
        if not self._queued_jobs:
            logger.debug("queue is empty; nothing to do")
            return

        jobs_to_pop = []
        available_jobs = self._queue_depth - len(self._outstanding_jobs)
        if available_jobs == 0:
            logger.debug("queue is full")
            return

        num_blocked = 0
        for i, job in enumerate(self._queued_jobs):
            blocking = job.get_blocking_jobs()
            if blocking:
                logger.debug("job %s is blocked by %s", job.name, blocking)
                num_blocked += 1
                continue

            self._run_job(job)
            jobs_to_pop.append(i)
            if len(jobs_to_pop) >= available_jobs:
                break

        for index in reversed(jobs_to_pop):
            self._queued_jobs.pop(index)

        logger.debug(
            "Started %s jobs in process_queue; num_blocked=%s", len(jobs_to_pop), num_blocked
        )

    def run(self, jobs):
        """
        Run job queue synchronously. Blocks until all jobs are complete.

        Parameters
        ----------
        jobs : list
            List of AsyncJobInterface objects to run.

        """
        for job in jobs:
            self.submit(job)

        self.wait()

    def submit(self, job):
        """Submit a job to be executed. If the queue is not full then it will
        run the job. Otherwise, it will queue the job. Returns immediately.
        The caller should call :meth:`JobQueue.process_queue` periodically to check
        for completions and start new jobs.

        Parameters
        ----------
        job : AsyncJobInterface

        """
        if self.is_full():
            logger.debug("queue depth exceeded, queue job %s", job.name)
            self._queued_jobs.append(job)
        elif job.get_blocking_jobs():
            logger.debug("Job is blocked by %s", job.get_blocking_jobs())
            self._queued_jobs.append(job)
        else:
            self._run_job(job)
            logger.debug("Started job %s in submit", job.name)

    def wait(self):
        """Return once all jobs have completed."""
        while self._outstanding_jobs or self._queued_jobs:
            self.process_queue()
            time.sleep(self._poll_interval)

        assert self._num_completed == self._num_jobs, f"{self._num_completed} {self._num_jobs}"

        self._handle_monitor_func(force=True)

    @classmethod
    def run_jobs(
        cls,
        jobs,
        max_queue_depth,
        poll_interval=DEFAULT_POLL_INTERVAL,
        monitor_func=None,
        monitor_interval=DEFAULT_MONITOR_INTERVAL,
    ):
        """
        Run job queue synchronously. Blocks until all jobs are complete.

        Parameters
        ----------
        jobs : list
            List of AsyncJobInterface objects
        max_queue_depth : int
            Maximum number of parallel jobs to maintain
        poll_interval : int
            Seconds to sleep in between completion checks.
        monitor_func : callable
            Optionally a function to call each poll_interval, such as for
            resource monitoring.
        monitor_interval : int
            Interval in seconds on which to run monitor_func.

        """
        queue = cls(
            max_queue_depth,
            poll_interval=poll_interval,
            monitor_func=monitor_func,
            monitor_interval=monitor_interval,
        )
        queue.run(jobs)



"""CLI to run jobs."""

import logging
import os
import sys
import time

import click

from jade.jobs.cluster import Cluster
from jade.jobs.job_submitter import JobSubmitter
from jade.loggers import setup_logging
from jade.utils.run_command import run_command
from jade.utils.utils import get_cli_string


logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    "output",
    type=click.Path(exists=True),
)
@click.option(
    "--complete/--no-complete",
    default=True,
    is_flag=True,
    show_default=True,
    help="Run completion operations. This can take some time. Use --no-complete if you plan to "
    "discard the results.",
)
@click.option(
    "--verbose", is_flag=True, default=False, show_default=True, help="Enable verbose log output."
)
def cancel_jobs(output, complete, verbose):
    """Cancels jobs."""
    filename = os.path.join(output, "cancel_jobs.log")
    level = logging.DEBUG if verbose else logging.INFO
    setup_logging(__name__, filename, file_level=level, console_level=level, mode="a")
    logger.info(get_cli_string())

    for _ in range(60):
        cluster, promoted = Cluster.deserialize(
            output,
            try_promote_to_submitter=True,
            deserialize_jobs=True,
        )
        if not promoted:
            logger.info("Did not get promoted. Sleep.")
            time.sleep(1)
            continue
        if cluster.is_complete():
            cluster.demote_from_submitter()
            logger.info("All jobs are already finished.")
            sys.exit(0)
        submitter = JobSubmitter.load(output)
        submitter.cancel_jobs(cluster)
        cluster.demote_from_submitter()
        ret = 0
        if complete:
            delay = 15
            print(f"Delaying {delay} seconds to let the nodes complete.")
            time.sleep(delay)
            ret = run_command(f"jade try-submit-jobs {output}")
        sys.exit(ret)

    logger.error("Failed to get promoted to submitter.")
    sys.exit(1)



"""Generates metrics from a Spark cluster."""

import logging
import requests
from pathlib import Path

from jade.exceptions import ExecutionError
from jade.utils.utils import dump_data


logger = logging.getLogger(__name__)


class SparkMetrics:
    """Records metrics from a Spark cluster"""

    METRICS_FILE = "metrics.json"

    def __init__(self, manager_node, history=False):
        self._manager_node = manager_node
        port = "18080" if history else "4040"
        self._endpoint = f"http://{manager_node}:{port}/api/v1/applications/"

    def _submit_request(self, cmd, *args):
        if not cmd.endswith("/"):
            cmd += "/"
        if args:
            cmd = cmd + "/".join(args)
        logger.info("Submitting %s", cmd)
        response = requests.get(cmd)
        if response.status_code != 200:
            raise ExecutionError(f"{cmd} failed: status_code={response.status_code}")

        return response.json()

    def generate_metrics(self, output_dir: Path):
        """Generate metrics from a Spark cluster into files at the given path.

        Parameters
        ----------
        cluster: str
            Cluster master node name
        output_dir: Path
            Output directory in which to write files

        """
        output_dir.mkdir(exist_ok=True, parents=True)
        apps = self.list_applications()
        results = {"metrics": []}
        for app in apps:
            metrics = {
                "application": app,
                "executors": self._submit_request(self._endpoint, app["id"], "executors"),
                "jobs": self._submit_request(self._endpoint, app["id"], "jobs"),
            }
            results["metrics"].append(metrics)

        filename = output_dir / self.METRICS_FILE
        dump_data(results, filename, indent=2)
        logger.info("Recorded metrics in %s", filename)

    def list_applications(self):
        """Return the applications in the cluster.

        Returns
        -------
        list
            List of application objects

        """
        return self._submit_request(self._endpoint)
